import { NextFunction, Request, Response } from 'express';
import { AddressInfo } from 'net';
import * as path from 'path';
import initRoutes from '../routes';
import { logger, reportError, sentry } from './utils';

import bodyParser = require('body-parser');
import cookieParser = require('cookie-parser');
import express = require('express');
import http = require('http');
import cors = require('cors');

const uuidv4: any = require('uuid/v4');

const devMode = process.env.APP_ENV === 'development';

const CORS_WHITELIST = [
  'http://localhost:4444',
  'http://localhost:7777',
  'https://m.scpfoundation.net',
  'https://lunory.scpfoundation.net',
  'https://resure.scpfoundation.net',
];

const httpLogger = require('pino-http')({
  logger,
  genReqId: uuidv4,
  serializers: {
    req: (req: Request) => {
      if (devMode) {
        return `${req.method} ${req.url} [${req.id}]`;
      }
      return {
        id: req.id,
        method: req.method,
        url: req.url,
        headers: req.headers,
        remoteAddress: req.connection && req.connection.remoteAddress,
        remotePort: req.connection && req.connection.remotePort,
      };
    },
    res: (res: Response) => {
      if (devMode) {
        const cleanedHeaders = String(res.header)
          .replace(/(\r\n)+/g, ' • ')
          .replace(/\s•\s$/, '');
        return `[${res.statusCode}] ${cleanedHeaders}`;
      }
      return {
        statusCode: res.statusCode,
        header: res.header,
      };
    },
  },
});

export default function generateServer(): (port: string | number) => void {
  const app = express();
  const router = express.Router(); // eslint-disable-line new-cap

  app.disable('x-powered-by');
  app.use(httpLogger);
  app.use(sentry.requestHandler());
  app.use(bodyParser.json());
  app.use(cors({
    origin: CORS_WHITELIST,
  }));

  app.use(cookieParser());

  initRoutes(app);
  app.use('/', router);

  app.use(sentry.errorHandler());

  app.use((err: Error, req: Request, res: Response, next: NextFunction) => {
    if (req.log) {
      req.log.error(err, 'Unhandled error');
    } else {
      reportError(err, { message: 'Unhandled error' });
    }
    res.status(500).send('Internal server error');
  });

  return function listen(port: string | number): void {
    const server = http.createServer(app);
    server.on('listening', () => {
      logger.info(`Express server is listening on port ${(server.address() as AddressInfo).port}`);
    });
    server.on('error', (error) => {
      if ((error as any).syscall !== 'listen') {
        throw error;
      }
      switch ((error as any).code) {
        case 'EACCES':
          logger.error(`Port ${port} requires elevated privileges`);
          process.exit(1);
          break;
        case 'EADDRINUSE':
          logger.error(`Port ${port} is already in use`);
          process.exit(1);
          break;
        default:
          throw error;
      }
    });
    server.listen(port);
  };
}
